import secrets
import string

from django.core.management.base import BaseCommand
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User, Permission, Group


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('username')
        parser.add_argument('email')

    def handle(self, *args, **options):
        user, created = User.objects.update_or_create(
            {'is_staff': True, 'email': options['email']}, username=options['username'])
        alphabet = string.ascii_letters + string.digits
        password = ''.join(secrets.choice(alphabet) for i in range(15))
        user.set_password(password)
        group, _ = Group.objects.get_or_create(name='TaskAdmins')
        group.permissions.add(*Permission.objects.filter(
            content_type__in=ContentType.objects.filter(app_label='django_tasks')))
        group.permissions.add(*Permission.objects.filter(
            content_type__in=ContentType.objects.filter(app_label='authtoken'),
            name='Can add Token'))
        user.groups.add(group)
        user.save()
        self.stdout.write(("Created user " if created else "Updated existing user ") +
                          f"{user} with email {options['email']} and NEW password, "
                          f"belonging to the {group} group.")
        return password



import asyncio
import collections
import importlib
import json

import bs4
import pytest
import pytest_asyncio

from channels.testing import WebsocketCommunicator
from django.core.management import call_command
from django.test.client import AsyncClient
from rest_framework import status

from bdd_coder import decorators
from bdd_coder import tester

from django_tasks.task_runner import TaskRunner


@pytest.mark.django_db
class BddTester(tester.BddTester):
    """
    The BddTester subclass of this tester package.
    It manages scenario runs. All test classes inherit from this one,
    so generic test methods for this package are expected to be defined here
    """
    gherkin = decorators.Gherkin(logs_path='bdd_runs.log')
    runner = TaskRunner.get()

    task_durations = [0.995, 0.95, 0.94, 0.8]
    credentials = dict(username='Alice', password='AlicePassWd')

    @pytest.fixture(autouse=True)
    def setup_clients(self):
        self.admin_client = AsyncClient()
        self.api_client = AsyncClient()

    @pytest.fixture(autouse=True)
    def setup_models(self):
        from django_tasks import models
        self.models = models

    async def assert_rest_api_call(self, method, api_path, expected_http_code, **kwargs):
        if 'json_data' in kwargs:
            kwargs['content_type'] = 'application/json'
            kwargs['data'] = json.dumps(kwargs.pop('json_data'))

        kwargs.setdefault('HTTP_AUTHORIZATION', 'Token ' + self.get_output('token'))
        response = await getattr(self.api_client, method.lower())(f'/api/{api_path}', **kwargs)
        assert response.status_code == expected_http_code, response.content.decode()
        return response

    @pytest_asyncio.fixture(autouse=True)
    async def setup_websocket_communicator(self, event_loop):
        from django_tasks import asgi
        route = asgi.application.application_mapping['websocket'].application.routes[0]
        consumers = importlib.import_module(route.lookup_str.rsplit('.', 1)[0])
        self.communicator = WebsocketCommunicator(
            consumers.TaskEventsConsumer.as_asgi(), route.pattern.describe().strip("'"))

        connected, subprotocol = await self.communicator.connect()
        assert connected

        self.event_collection_task = asyncio.wrap_future(
            asyncio.run_coroutine_threadsafe(self.collect_events(), event_loop))

        yield
        await self.communicator.disconnect()

    async def collect_events(self):
        self.events = collections.defaultdict(list)
        listen = True
        while listen:
            try:
                event = await self.communicator.receive_json_from(timeout=5)
            except asyncio.TimeoutError:
                listen = False
            else:
                self.events[event['content']['status'].lower()].append(event)

    async def fake_task_coro_ok(self, duration):
        await asyncio.sleep(duration)
        return duration

    async def fake_task_coro_raise(self, duration):
        await asyncio.sleep(duration)
        raise Exception('Fake error')

    def get_all_admin_messages(self, soup):
        return {k: self.get_admin_messages(soup, k) for k in ('success', 'warning', 'info')}

    @staticmethod
    def get_admin_messages(soup, message_class):
        return [li.contents[0] for li in soup.find_all('li', {'class': message_class})]

    @staticmethod
    def get_soup(response):
        return bs4.BeautifulSoup(response.content.decode(), features='html.parser')

    def a_tasks_admin_user_is_created_with_command(self, django_user_model):
        self.credentials['password'] = call_command(
            'create_task_admin', self.credentials['username'], 'fake@gmail.com')
        self.assert_login()
        user = django_user_model.objects.get(username=self.credentials['username'])
        return user,

    def assert_login(self):
        accepted = self.admin_client.login(**self.credentials)
        assert accepted, self.credentials

    async def assert_admin_redirection(self, response):
        assert response.status_code == status.HTTP_302_FOUND
        redirected_response = await self.admin_client.get(response.headers['Location'])
        assert redirected_response.status_code == status.HTTP_200_OK
        return redirected_response

    async def cancelled_error_success_messages_are_broadcasted(self):
        cancelled, error, success = map(int, self.param)
        await self.event_collection_task
        assert len(self.events['started']) == cancelled + error + success
        assert len(self.events['cancelled']) == cancelled
        assert len(self.events['error']) == error
        assert len(self.events['success']) == success



import os

from typing import Any

from django_tasks.settings import SettingsIni

CHANNEL_TASKS = SettingsIni()

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = 'cg2fbx3f)_9znm3$($suorm*0fyuv#wr586195!q^pv0%ct7c5'

DEBUG = CHANNEL_TASKS.debug

ALLOWED_HOSTS: list[str] = CHANNEL_TASKS.allowed_hosts

INSTALLED_APPS: list[str] = [
    'bootstrap5',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'rest_framework.authtoken',
    'rest_framework',
    'django.contrib.messages',
    'django_tasks.apps.TasksConfig',
    'django_extensions',
    'django_filters',
    'django.contrib.postgres',
]

MIDDLEWARE: list[str] = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'request_logging.middleware.LoggingMiddleware',
]

DJANGO_LOG_LEVEL = CHANNEL_TASKS.log_level

LOGGING = dict(
    version=1,
    disable_existing_loggers=False,
    handlers={
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'thread-logname',
        },
        'console-debug': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
    },
    formatters={
        'verbose': {
            'format': '{levelname} {asctime} {threadName} ({pathname}) {funcName}:L{lineno} ★ {message}',
            'style': '{',
        },
        'thread-logname': {
            'format': '{levelname} {asctime} ({threadName}) {name} ★ {message}',
            'style': '{',
        },
    },
    loggers={
        'django': {
            'level': DJANGO_LOG_LEVEL,
            'handlers': ['console-debug'],
        },
        'django.request': {
            'handlers': ['console'],
            'level': DJANGO_LOG_LEVEL,
            'propagate': False,
        },
        'django.channels': {
            'handlers': ['console'],
            'level': DJANGO_LOG_LEVEL,
            'propagate': False,
        },
    },
)

ROOT_URLCONF = 'django_tasks.urls'

TEMPLATES: list[dict[str, Any]] = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

ASGI_APPLICATION = 'django_tasks.ws_asgi.application'

DATABASES: dict[str, Any] = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'postgres',
        'USER': 'postgres',
        'PASSWORD': 'postgres',
        'HOST': 'db',
        'PORT': 5433,
    }
}

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

AUTHENTICATION_BACKENDS: list[str] = [
    "django.contrib.auth.backends.ModelBackend",
]

SESAME_MAX_AGE = 86400*5

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Europe/Madrid'

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')
MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

REST_FRAMEWORK = dict(
    DEFAULT_RENDERER_CLASSES=(
        'rest_framework.renderers.JSONRenderer',
    ),
    DEFAULT_PAGINATION_CLASS='rest_framework.pagination.PageNumberPagination',
    DEFAULT_FILTER_BACKENDS=('django_filters.rest_framework.DjangoFilterBackend',),
    DEFAULT_AUTHENTICATION_CLASSES=(
        'rest_framework.authentication.TokenAuthentication',
    ),
    DEFAULT_PERMISSION_CLASSES=(
        'rest_framework.permissions.IsAuthenticated',
    ),
    TEST_REQUEST_DEFAULT_FORMAT='json',
)

CHANNEL_LAYERS = dict(default={
    "BACKEND": "channels_redis.core.RedisChannelLayer",
    "CONFIG": {
        "hosts": [("redis", 6379)],
    },
})



import asyncio

from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from rest_framework import status

from django_tasks.drf_consumer import DrfConsumer
from django_tasks.doctask_scheduler import DocTaskScheduler
from django_tasks.serializers import DocTaskSerializer


class TasksRestConsumer(DrfConsumer, DocTaskScheduler):
    async def process_drf_response(self, drf_response):
        if drf_response.status_code == status.HTTP_201_CREATED:
            if isinstance(drf_response.data, dict):
                await self.schedule_doctask(drf_response.data)
            elif isinstance(drf_response.data, list):
                await asyncio.gather(*[self.schedule_doctask(data) for data in drf_response.data])


class TaskEventsConsumer(AsyncJsonWebsocketConsumer, DocTaskScheduler):
    groups = ['tasks']

    async def task_started(self, event):
        """Echoes the task.started document."""
        await self.handle_doctask(event['content'])
        await self.send_json(content=event)

    async def task_success(self, event):
        """Echoes the task.success document."""
        await self.send_json(content=event)

    async def task_cancelled(self, event):
        """Echoes the task.cancelled document."""
        await self.send_json(content=event)

    async def task_error(self, event):
        """Echoes the task.error document."""
        await self.send_json(content=event)

    async def receive_json(self, content):
        """Pocesses task schedule websocket requests."""
        await DocTaskSerializer.schedule_task_group(content)

    async def handle_doctask(self, event_content):
        doctask = await database_sync_to_async(self.retrieve_doctask)(event_content['memory-id'])
        if doctask:
            doctask.document.append(event_content)
            await doctask.asave()



import datetime
import json

from django.db.models import Model, CharField, DateTimeField, JSONField


class DefensiveJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return super().default(obj)
        except TypeError:
            return repr(obj)


class DocTask(Model):
    """Stored information of a task execution."""
    name: CharField = CharField(max_length=80)
    scheduled_at: DateTimeField = DateTimeField(default=datetime.datetime.now)
    completed_at: DateTimeField = DateTimeField(null=True)
    inputs: JSONField = JSONField(default=dict, encoder=DefensiveJsonEncoder)
    document: JSONField = JSONField(default=list, encoder=DefensiveJsonEncoder)

    def __str__(self):
        return f'Doc-task {self.pk}, ' + (
            f'completed at {self.completed_at}, took {self.duration}' if self.completed_at
            else f'running for {self.duration}')

    @property
    def duration(self):
        return (self.completed_at if self.completed_at else datetime.datetime.now()) - self.scheduled_at

    async def on_completion(self, task_info):
        self.completed_at = datetime.datetime.now()
        self.document.append(task_info)
        await self.asave()



from rest_framework import decorators, response, status, viewsets

from django_tasks import models, serializers


class TaskViewSet(viewsets.ModelViewSet):
    http_method_names = ['get', 'post', 'delete', 'head', 'options', 'trace']
    queryset = models.DocTask.objects.all()
    serializer_class = serializers.DocTaskSerializer

    @decorators.action(detail=False, methods=['post'])
    def schedule(self, request, *args, **kwargs):
        """DRF action that schedules an array of tasks."""
        many_serializer, _ = self.serializer_class.create_doctask_group(
            request.data, context=self.get_serializer_context())

        return response.Response(data=many_serializer.data, status=status.HTTP_201_CREATED)



from __future__ import annotations

import asyncio
import logging
import threading
import traceback

from typing import Any, Callable, Coroutine

from channels.layers import get_channel_layer


class TaskRunner:
    """
    Class in charge of in-memory handling of `asyncio` background tasks, with a worker thread per instance.
    """
    _instances: list[TaskRunner] = []

    @classmethod
    def get(cls) -> TaskRunner:
        """
        Returns the last instance created, a new one if necessary, and ensures that its worker thread is alive.
        """
        if not cls._instances:
            cls()

        cls._instances[-1].ensure_alive()

        logging.getLogger('django').debug('Using task runner: %s.', cls._instances[-1])
        return cls._instances[-1]

    def __init__(self):
        self.event_loop = asyncio.new_event_loop()
        self.worker_thread = threading.Thread(target=self.event_loop.run_forever, daemon=True)
        self.__class__._instances.append(self)
        logging.getLogger('django').debug('New task runner: %s.', self)

    def __str__(self):
        return f'loop={self.event_loop}, worker={self.worker_thread}'

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self}>'

    def ensure_alive(self):
        """Ensures the worker thread is alive."""
        if not self.worker_thread.is_alive():
            self.worker_thread.start()

    def run_coroutine(self, coroutine: Coroutine) -> asyncio.Future:
        """Runs the given `coroutine` thread-safe in the loop."""
        return asyncio.wrap_future(asyncio.run_coroutine_threadsafe(coroutine, self.event_loop))

    def run_on_task_info(self,
                         async_callback: Callable[[dict[str, Any]], Coroutine],
                         task: asyncio.Future) -> asyncio.Future:
        """Runs the `async_callback` taking the task info as the argument."""
        return self.run_coroutine(async_callback(self.get_task_info(task)))

    async def schedule(self,
                       coroutine: Coroutine,
                       *coro_callbacks: Callable[[dict[str, Any]], Coroutine]) -> asyncio.Future:
        task_name = coroutine.__name__
        task = self.run_coroutine(coroutine)
        await self.broadcast_task(task_name, task)

        task.add_done_callback(lambda tk: self.run_coroutine(self.broadcast_task(task_name, tk)))

        for coro_callback in coro_callbacks:
            task.add_done_callback(lambda tk: self.run_on_task_info(coro_callback, tk))

        return task

    @classmethod
    async def broadcast_task(cls, name: str, task: asyncio.Future):
        """
        Sends the task info to the 'tasks' group of consumers, specifying a message type per task status.
        """
        task_info = cls.get_task_info(task)
        task_info['name'] = name
        message_type = f"task.{task_info['status'].lower()}"
        channel_layer = get_channel_layer()
        await channel_layer.group_send("tasks", {'type': message_type, 'content': task_info})

    @staticmethod
    def get_task_info(task: asyncio.Future) -> dict[str, Any]:
        """Extracts and returns the corresponding task info, with the current task status."""
        task_info: dict[str, Any] = {'memory-id': id(task)}

        if not task.done():
            task_info['status'] = 'Started'
        elif task.cancelled():
            task_info['status'] = 'Cancelled'
        elif task.exception():
            task_info.update({'status': 'Error',
                              'exception-type': task.exception().__class__.__name__,
                              'traceback': traceback.format_exception(task.exception())})
        else:
            task_info.update({'status': 'Success', 'output': task.result()})

        return task_info



import asyncio
import logging

from django_tasks import models
from django_tasks.task_inspector import TaskCoroInfo


@TaskCoroInfo.mark_as_safe
async def sleep_test(duration, raise_error=False):
    logging.getLogger('django').info('Starting sleep test.')
    await asyncio.sleep(duration)

    if raise_error:
        logging.getLogger('django').info('Sleep test done with raise.')
        raise Exception('Test error')

    logging.getLogger('django').info('Sleep test done with no raise.')
    return f"Slept for {duration} seconds"


@TaskCoroInfo.mark_as_safe
async def doctask_access_test(instance_ids: list[int]):
    await asyncio.sleep(3)
    async for doctask in models.DocTask.objects.filter(pk__in=instance_ids):
        logging.getLogger('django').info('Retrieved %s', repr(doctask))
    await asyncio.sleep(3)


@TaskCoroInfo.mark_as_safe
async def doctask_deletion_test(instance_ids: list[int]):
    await asyncio.sleep(3)
    await models.DocTask.objects.filter(pk__in=instance_ids).adelete()
    logging.getLogger('django').info('Deleted doctasks %s', instance_ids)
    await asyncio.sleep(3)



import traceback

from channels.db import database_sync_to_async
from channels.generic.http import AsyncHttpConsumer

from django import urls
from django.conf import settings

from rest_framework import status
from rest_framework.test import APIRequestFactory


class DrfConsumer(AsyncHttpConsumer):
    content_type = 'application/json'

    def __init__(self, *args, **kwargs):
        self.drf_url = kwargs.pop('drf_url')
        super().__init__(*args, **kwargs)

    @classmethod
    def get_urls(cls, drf_router):
        return [urls.re_path(str(url.pattern), cls.as_asgi(drf_url=url)) for url in drf_router.urls]

    @property
    def args(self):
        return self.scope['url_route']['args']

    @property
    def kwargs(self):
        return self.scope['url_route']['kwargs']

    def make_drf_request(self, request_body: bytes):
        factory = APIRequestFactory()
        return getattr(factory, self.scope['method'].lower())(
            self.scope['path'],
            request_body,
            headers={k.decode(): v.decode() for k, v in self.scope['headers']},
            content_type=self.content_type,
        )

    @database_sync_to_async
    def view_coroutine(self, request_body: bytes):
        return self.drf_url.callback(self.make_drf_request(request_body), *self.args, **self.kwargs)

    async def handle(self, body: bytes):
        try:
            drf_response = await self.view_coroutine(body)
            drf_response.render()
            await self.process_drf_response(drf_response)
            await self.send_response(drf_response.status_code, drf_response.content, headers=[
                (k.encode(), v.encode()) for k, v in drf_response.headers.items()
            ])
        except Exception as exc:
            await self.send_response(
                status.HTTP_500_INTERNAL_SERVER_ERROR,
                ''.join(traceback.format_exception(exc)).encode() if settings.DEBUG else b''
            )

    async def process_drf_response(self, drf_response):
        """Overrite for any postprocessing."""
